(*---------------------------------------------------------------------------*
  $Change$
  Copyright (c) 2006-2010, James H. Woodyatt
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  
    Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
  COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGE. 
 *---------------------------------------------------------------------------*)

(** Connection-oriented, orderly close (SOCK_STREAM) sockets. *)

(** {6 Overview}
    
    TCP and other SOCK_STREAM sockets are connection-oriented communications
    channels with reliable delivery of octet streams and orderly close
    semantics.  Classes and functors are defined to extend the core socket
    functions in {!Iom_socket} to add these features.
*)

(** An I/O stream event representing a connection event. *)
class ['af] connection:
    Iom_stream.more -> ('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t ->
    object
        (** Flag to indicate whether more connections are expected. *)
        method more: Iom_stream.more
        
        (** The socket descriptor. *)
        method socket: ('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t
    end

(** A listener has started at the specified address. *)
type 'address listen = [ `Listen of 'address ]

(** The notifications from listeners. *)
type 'address listener_notify = [ 'address listen | Iom_stream.failed ]

(** {6 Simplified Functional Interface} *)

(** The fix for controlling a socket. *)
type endpoint_fix =
    (Iom_octet_stream.fragment, Iom_octet_stream.fragment,
        Iom_stream.flowcontrol, Iom_stream.flownotify) Iom_stream.iofix

(** The fix for controlling a connection initiator. *)
type 'af initiator_fix =
    ('af connection, Iom_stream.flowcontrol, Iom_stream.failed)
        Iom_stream.ifix

(** The fix for controlling a listener. *)
type 'af listener_fix =
    ('af connection, Iom_stream.flowcontrol, 'af Nx_socket.sockaddr
    listener_notify) Iom_stream.ifix

(** Evaluate [initiator ~s ~src ~dst k] to obtain the initiator fix for
    binding the socket [s] to the local address [src] and connecting it in the
    kernel context [k] to the destination address [dst].
*)
val initiator:
    s:('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t ->
    src:'af Nx_socket.sockaddr -> dst:'af Nx_socket.sockaddr ->
    Iom_gadget.kernel -> 'af initiator_fix Iom_gadget.t

(** Evaluate [listener ~s ~src ~backlog k] to obtain the listener fix for
    binding the socket [s] to the local address [src], applying a listen queue
    depth of [backlog] and accepting connections in the kernel context [k].
*)
val listener:
    s:('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t -> 
    src:'af Nx_socket.sockaddr -> backlog:int -> Iom_gadget.kernel ->
    'af listener_fix Iom_gadget.t

(** Evaluate [endpoint ~s ?addrs ~limits k] to obtain the endpoint fix for
    the socket [s].  If [~addrs] is provided, then bind the socket to the
    first address and connect to the second address before notifying ready to
    send.
*)
val endpoint:
    s:('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t ->
    ?addrs:('af Nx_socket.sockaddr * 'af Nx_socket.sockaddr) ->
    limits:Iom_octet_stream.limits -> Iom_gadget.kernel ->
    endpoint_fix Iom_gadget.t

(** {6 Signaling Connection} *)

(** Use [inherit connector ~k ~s ~rwx pad] to derive a subclass gadget that
    responds to connection events on the socket [s] in the kernel context [k].
    The [rwx] parameter identifies the sort of {!Nx_poll.file} required.  The
    base class inherits from {!Iom_reactor.source} and {!Nx_poll.file}.
*)
class virtual ['control, 'notify, 'connect] connector:
    k:Iom_gadget.kernel -> s:('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t ->
    rwx:[< Nx_poll.rwx ] -> ('connect, 'control, 'notify) Iom_stream.ipad ->
    object
        inherit ['control, 'connect] Iom_reactor.source
        inherit [unit Iom_gadget.t] Nx_poll.file
        
        constraint 'control = [> Iom_stream.flowcontrol ]
        constraint 'notify = [> Iom_stream.failed ]
        constraint 'connect = 'af #connection
        
        (** Set [true] when the gadget is started.  Because reactor state is
            mutable, attempts to start more than one gadget with the same
            initial state object raises [Failure].
        *)
        val mutable initiated_: bool
        
        (** Method called during [self#start] to bind the socket to its local
            address.
        *)
        method virtual private bind: unit
        
        (** Use [self#fail x] to close the connector socket, notify [`Failed
            x] and terminate the gadget.
        *)
        method private fail: 'a. exn -> 'a Iom_gadget.t
    end

(** Use [inherit \['control, 'notify, 'connect\] initiator ~k ~s pad] to
    derive a subclass for a gadget that initiates a connection to a remote
    address in the kernel context [k] with the socket [s].
*)
class virtual ['control, 'notify, 'connect] initiator:
    k:Iom_gadget.kernel -> s:('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t ->
    ('connect, 'control, 'notify) Iom_stream.ipad ->
    object
        inherit ['control, 'notify, 'connect] connector
        constraint 'connect = 'af #connection
        
        (** Method called during [self#source] to construct the connection
            object.
        *)
        method virtual private connect: 'connect
        
        (** Returns [Some self#connect]. *)
        method private source: 'connect option
    end

(** Use [inherit \['control, 'notify, 'connect\] listener ~k ~s ~backlog pad]
    to derive a subclass for a gadget that listens for connections at a local
    address in the kernel context [k] with the socket [s].
*)
class virtual ['control, 'notify, 'connect] listener:
    k:Iom_gadget.kernel -> s:('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t ->
    backlog:int -> ('connect, 'control, 'notify) Iom_stream.ipad ->
    object
        inherit ['control, 'notify, 'connect] connector
        constraint 'connect = 'af #connection
        constraint 'notify = [> 'address listener_notify ]
        
        (** The list of pending connections accepted but not yet drained. *)
        val mutable stack_: 'connect list
        
        (** Define [self#getsockname] to return the local address of the
            listening socket.
        *)
        method virtual private getsockname: 'address
        
        (** Define [self#accept] to accept a new connection and construct the
            associated [connection] object for it.
        *)
        method virtual private accept: 'connect
    end

(** {6 Endpoint Control} *)

(** A receiver gadget for stream socket input. *)
class ['control, 'notify] receiver:
    k:Iom_gadget.kernel -> s:('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t ->
    limits:Iom_octet_stream.limits ->
    (Iom_octet_stream.fragment, 'control, [> `Finished ] as 'notify)
        Iom_stream.ipad ->
    ['control, 'notify] Iom_reactor.reader

(** A sender gadget for stream socket output. *)
class ['control, 'notify] sender:
    k:Iom_gadget.kernel -> s:('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t ->
    (Iom_octet_stream.fragment, 'control, [> `Finished ] as 'notify)
        Iom_stream.opad ->
    [Iom_octet_stream.fragment, 'control, 'notify] Iom_reactor.writer

(** The endpoint control state for stream socket gadgets.  Extends the core
    socket state type to add connection-oriented semantics.
*)
type ('ci, 'co, 'ni, 'no, 'x, 'cx, 'nx) state = [

    (** Core states. *)
    | ('ci, 'co, 'ni, 'no) Iom_socket.state
    
    (** Socket is receiving only.  Sender has stopped. *)
    | `S_receiving of ('ci, 'ni) Iom_gadget.fix
    
    (** Socket is sending only.  Receiver has stopped. *)
    | `S_sending of ('co, 'no) Iom_gadget.fix
    
    (** Socket is initiating a connection. *)
    | `S_initiating of ('x, 'cx, 'nx) Iom_stream.ifix * 'ci
]
constraint 'ni = [> `Finished ]
constraint 'no = [> `Finished ]

(** Use [inherit \['i, 'o, 'control, 'notify, 'af\] endpoint ~k ~s ?addrs pad]
    to derive an endpoint controller gadget for stream sockets specialized by
    the socket address family.
*)
class virtual ['i, 'o, 'control, 'notify, 'af] endpoint:
    k:Iom_gadget.kernel -> s:('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t ->
    ?addrs:('af Nx_socket.sockaddr * 'af Nx_socket.sockaddr) ->
    ('i, 'o, 'control, 'notify) Iom_stream.iopad ->
    object
        inherit ['i, 'o, 'control, 'notify, 'state, 'af, 'st]
            Iom_socket.endpoint
        
        constraint 'i = #Iom_octet_stream.fragment
        constraint 'o = #Iom_octet_stream.fragment
        constraint 'state = ('ci, 'co, 'ni, 'no, 'x, 'cx, 'nx) state
        constraint 'st = Nx_socket.SOCK_STREAM.tag
        
        (** Called in [self#state0] (unless [?addrs] is [None]) to construct
            an initiator gadget for the source and destination address with
            the pad provided.  The initiator is started and the state is set
            to [`S_initiating].
        *)
        method virtual private initiator:
            src:'af Nx_socket.sockaddr -> dst:'af Nx_socket.sockaddr ->
            ('x, 'cx, 'nx) Iom_stream.ipad -> #Iom_gadget.start
    end

(** {6 Specializer Functor}
    
    Use the interface defined below to specialize the stream socket gadget for
    a specific address family.
*)

(** Auxillary class types referenced in the module type [T] below. *)
module Aux: sig
    (** A synonym for the core [connection] class type. *)
    class type ['af] connection_aux = ['af] connection
    
    (** A subclass type of the core [initiator] class type that defines a
        virtual method [aux] to produce a [connection] object.
    *)
    class type virtual ['control, 'notify, 'connect] initiator_aux = object
        inherit ['control, 'notify, 'connect] initiator
        constraint 'connect = 'af #connection_aux
        
        (** Called in [self#connect] to produce a [connection] object. *)
        method virtual private aux:
            ('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t -> 'connect
    end
    
    (** A subclass type of the core [listener] class type that defines a
        virtual method [aux] to produce a [connection] object.
    *)
    class type virtual ['control, 'notify, 'connect] listener_aux = object
        inherit ['control, 'notify, 'connect] listener
        constraint 'connect = 'af #connection_aux
        
        method virtual private aux:
            ('af, Nx_socket.SOCK_STREAM.tag) Nx_socket.t -> 'connect
    end
    
    (** A synonym for the core [endpoint] class. *)
    class type virtual ['i, 'o, 'control, 'notify, 'af] endpoint_aux =
        ['i, 'o, 'control, 'notify, 'af] endpoint
end

(** The signature of modules produced by the [Create] functor. *)
module type T = sig
    include Iom_socket.T with module P.ST = Nx_socket.SOCK_STREAM
    
    (** The class of connection objects. *)
    class connection: Iom_stream.more -> t -> object
        inherit [P.AF.tag] Aux.connection_aux
        method local: address
        method remote: address
    end
    
    (** The fix for controlling a socket connection initator gadget. *)
    type initiator =
        (connection, Iom_stream.flowcontrol, Iom_stream.failed)
        Iom_stream.ifix
    
    (** The fix for controlling a socket connection listener gadget. *)
    type listener =
        (connection, Iom_stream.flowcontrol, address listener_notify)
        Iom_stream.ifix
    
    (** The fix for controlling a socket endpoint gadget. *)
    type endpoint = endpoint_fix
    
    (** Evaluate [initator ?s ?src ~dst k] to obtain the fix for a connection
        initiator to the destination [dst] in the kernel context [k].  If [src]
        is provided, then bind explicity to that local address.  If [s] is
        not provided, then construct a new socket.
    *)
    val initiator:
        ?s:t -> ?src:address -> dst:address -> Iom_gadget.kernel ->
        initiator Iom_gadget.t
    
    (** Evaluate [listener ?s ?src ~backlog k] to obtain the fix for a
        connection listener in the kernel context [k].  If [src] is not
        provided, then bind to an unspecified address.  If [s] is not provided,
        then construct a new socket.
    *)
    val listener:
        ?s:t -> ?src:address -> backlog:int -> Iom_gadget.kernel ->
        listener Iom_gadget.t
    
    (** Evaluate [endpoint ?s ?addrs ~limits k] to obtain the fix for an
        endpoint controller constrained on receive by [limits] in the kernel
        context [k].  If [addrs] is provided, then bind explicitly to the first
        address and connect to the second address before notifying ready to
        send.  If [s] is not provided, then construct a new socket.
    *)
    val endpoint:
        ?s:t -> ?addrs:(address * address) -> limits:Iom_octet_stream.limits ->
        Iom_gadget.kernel -> endpoint Iom_gadget.t
end

(** Use [Create(P)] to compose a module of stream socket gadgets specialized
    for the address family of the socket module [P].
*)
module Create(P: Nx_socket.P with module ST = Nx_socket.SOCK_STREAM):
    T with module P = P

(*--- $File$ ---*)
